<% content_for(:title, "Landing") %>

<div class="row">
  <div class="col-md col-lg-8 mx-auto pt-4">
    <div class="card mt-4">
      <div class="card-body">
        <div class="card-title">
          <h2>
            Welcome to the GraphQL-Ruby Dashboard
          </h2>
        </div>
        <p class="card-text">
          Click the links above to see data about your schema (<code><%= schema_class %></code>).
        </p>
      </div>
    </div>
  </div>
</div>
