# frozen_string_literal: true

<% module_namespacing_when_supported do -%>
module Types
  class BaseInputObject < GraphQL::Schema::InputObject
    argument_class Types::BaseArgument
  end
end
<% end -%>
