# frozen_string_literal: true

require "graphql/rubocop/graphql/default_null_true"
require "graphql/rubocop/graphql/default_required_true"
require "graphql/rubocop/graphql/field_type_in_block"
require "graphql/rubocop/graphql/root_types_in_block"
